from typing import NamedTuple


VERSION_TOML_URL = "https://www.batrachian.ai/toad.toml"


class VersionMeta(NamedTuple):
    """Information about the current version of Toad."""

    version: str
    upgrade_message: str
    visit_url: str


class VersionCheckFailed(Exception):
    """Something went wrong in the version check."""


async def check_version() -> tuple[bool, VersionMeta]:
    """Check for a new version of Toad.

    Returns:
        A tuple containing a boolean that indicates if there is a newer version,
            and a `VersionMeta` structure with meta information.
    """
    import httpx
    import packaging.version
    import tomllib

    from toad import get_version

    try:
        current_version = packaging.version.parse(get_version())
    except packaging.version.InvalidVersion as error:
        raise VersionCheckFailed(f"Invalid version;{error}")

    try:
        async with httpx.AsyncClient() as client:
            response = await client.get(VERSION_TOML_URL)
            version_toml_bytes = await response.aread()
    except Exception as error:
        raise VersionCheckFailed(f"Failed to retrieve version;{error}")

    try:
        version_toml = version_toml_bytes.decode("utf-8", "replace")
        version_meta = tomllib.loads(version_toml)
    except Exception as error:
        raise VersionCheckFailed(f"Failed to decode version TOML;{error}")

    if not isinstance(version_meta, dict):
        raise VersionCheckFailed("Response isn't TOML")

    toad_version = str(version_meta.get("version", "0"))
    version_message = str(version_meta.get("upgrade_message", ""))
    version_message = version_message.replace("$VERSION", toad_version)
    verison_meta = VersionMeta(
        version=toad_version,
        upgrade_message=version_message,
        visit_url=str(version_meta.get("visit_url", "")),
    )

    try:
        new_version = packaging.version.parse(verison_meta.version)
    except packaging.version.InvalidVersion as error:
        raise VersionCheckFailed(f"Invalid remote version;{error}")

    return new_version > current_version, verison_meta


if __name__ == "__main__":

    async def run() -> None:
        result = await check_version()
        from rich import print

        print(result)

    import asyncio

    asyncio.run(run())



from itertools import chain
from typing import Iterable, Sequence
from pathlib import Path
import pathspec
import pathspec.patterns
from pathspec import GitIgnoreSpec

import rich.repr


def load_path_spec(git_ignore_path: Path) -> GitIgnoreSpec | None:
    """Get a path spec instance if there is a .gitignore file present.

    Args:
        git_ignore_path): Path to .gitignore.

    Returns:
        A `PathSpec` instance.
    """
    try:
        if git_ignore_path.is_file():
            try:
                spec_text = git_ignore_path.read_text(encoding="utf-8")
            except Exception:
                # Permissions, encoding issue?
                return None
            try:
                spec = GitIgnoreSpec.from_lines(
                    pathspec.patterns.GitWildMatchPattern, spec_text.splitlines()
                )
            except Exception:
                return None
            return spec
    except OSError:
        return None
    return None


@rich.repr.auto
class PathFilter:
    """Filter paths according to .gitignore files."""

    def __init__(
        self, root: Path, path_specs: Iterable[GitIgnoreSpec] | None = None
    ) -> None:
        self._root = root
        self._default_specs = [] if path_specs is None else list(path_specs)
        self._path_specs: dict[Path, Sequence[GitIgnoreSpec]] = {}

    def __rich_repr__(self) -> rich.repr.Result:
        yield (str(self._root),)

    @classmethod
    def from_git_root(cls, path: Path) -> PathFilter:
        """Load all path specs from parent directories up to the most recent directory with .git

        Args:
            path: A directory path.

        Returns:
            PathFilter instance.
        """
        filter_root = path
        path_specs: list[GitIgnoreSpec] = []
        try:
            while (parent := path.parent) != parent:
                if (path_spec := load_path_spec(path / ".gitignore")) is not None:
                    path_specs.append(path_spec)
                if (path / ".git").is_dir():
                    break
                path = parent
            else:
                del path_specs[:]
        except OSError:
            pass
        return PathFilter(filter_root, reversed(path_specs))

    def get_path_specs(self, path: Path) -> Sequence[GitIgnoreSpec]:
        """Get a sequence of path specs applicable to the give path.

        This will inherit path specs up to the root path of the filtr.

        Args:
            path: A directory path

        Returns:
            A sequence of path specs.
        """
        if (cached_path_specs := self._path_specs.get(path)) is not None:
            return cached_path_specs
        path_spec = load_path_spec(path / ".gitignore")
        if path == self._root:
            path_specs = [path_spec] if path_spec is not None else []
        else:
            parent_path_specs = self.get_path_specs(path.parent)
            path_specs = (
                parent_path_specs
                if path_spec is None
                else [*parent_path_specs, path_spec]
            )
        self._path_specs[path] = path_specs
        return path_specs

    def match(self, path: Path) -> bool:
        """Match a path againt the path filter.

        Returns:
            `True` if the path should be removed, `False` if it should be included.
        """
        if path.name == ".git":
            return True
        path_specs = self.get_path_specs(path.parent)
        for path_spec in chain(self._default_specs, path_specs):
            if path_spec.match_file(path):
                return True
        return False


if __name__ == "__main__":
    path_filter = PathFilter.from_git_root(Path("."))

    for path in Path(".").iterdir():
        print(path_filter.match(path), path)
    print(path_filter)



from pygments.util import ClassNotFound
from pygments.lexers import get_lexer_by_name
from pygments.token import Token

SPECIAL = {Token.Name.Function.Magic, Token.Name.Function, Token.Name.Class}


def get_special_name_from_code(code: str, language: str) -> list[str]:
    try:
        lexer = get_lexer_by_name(
            language,
            stripnl=False,
            ensurenl=True,
            tabsize=8,
        )
    except ClassNotFound:
        lexer = get_lexer_by_name(
            "text",
            stripnl=False,
            ensurenl=True,
            tabsize=8,
        )
    special: list[str] = []
    for token_type, token in lexer.get_tokens(code):
        if token_type in SPECIAL:
            special.append(token)
    return special



from textual.widgets import Markdown
from textual.widgets._markdown import MarkdownBlock
from textual.content import Content


class ConversationCodeFence(Markdown.BLOCKS["fence"]):
    pass


CUSTOM_BLOCKS = {"fence": ConversationCodeFence}


class ConversationMarkdown(Markdown):
    """Markdown widget with custom blocks."""

    def get_block_class(self, block_name: str) -> type[MarkdownBlock]:
        if (custom_block := CUSTOM_BLOCKS.get("block_name")) is not None:
            return custom_block
        return super().get_block_class(block_name)



from functools import lru_cache
import os.path
from typing import Iterable

from rich.cells import cell_len
from textual.geometry import Size
from textual.reactive import reactive
from textual.content import Content
from textual.widget import Widget


def radiate_range(total: int) -> Iterable[tuple[int, int]]:
    """Generate pairs of indexes, gradually growing from the center.

    Args:
        total: Total size of range.

    Yields:
        Pairs of indexes.
    """
    if not total:
        return
    left = right = total // 2
    yield (left, right)
    while left >= 0 or right < total:
        left -= 1
        if left >= 0:
            yield (left + 1, right)
        right += 1
        if right <= total:
            yield (left + 1, right)


@lru_cache(maxsize=16)
def condense_path(path: str, width: int, *, prefix: str = "") -> str:
    """Condense a path to fit within the given cell width.

    Args:
        path: The path to condense.
        width: Maximum cell width.
        prefix: A string to be prepended to the result.

    Returns:
        A condensed string.
    """
    # TODO: handle OS separators and path issues
    if cell_len(path) <= width:
        return path
    components = path.split("/")
    condensed = components
    trailing_slash = path.endswith("/")
    candidate = prefix + "/".join(condensed)
    if trailing_slash and candidate and not candidate.endswith("/"):
        candidate += "/"

    for left, right in radiate_range(len(components)):
        if cell_len(candidate) < width:
            return candidate
        condensed = [*components[:left], "…", *components[right:]]
        candidate = prefix + "/".join(condensed)
        if trailing_slash and candidate and not candidate.endswith("/"):
            candidate += "/"

    return candidate


class CondensedPath(Widget):
    path = reactive("")
    display_path = reactive("")

    DEFAULT_CSS = """
    CondensedPath {
        height: 1;
    }
    """

    def __init__(
        self,
        path: str = "",
        *,
        directory: bool = False,
        id: str | None = None,
        classes: str | None = None
    ) -> None:
        super().__init__(id=id, classes=classes)
        self.set_reactive(CondensedPath.path, path)
        self._directory = directory

    def on_resize(self) -> None:
        self.watch_path(self.path)

    def watch_path(self, path: str) -> None:
        if not path or not self.size.width:
            return
        path = os.path.abspath(os.path.expanduser(path))
        if not self.tooltip:
            self.tooltip = str(path)
        user_root = os.path.abspath(os.path.expanduser("~/"))
        if not user_root.endswith("/"):
            user_root += "/"
        if path.startswith(user_root):
            path = "~/" + path[len(user_root) :]
        if self._directory and not path.endswith("/"):
            path += "/"
        self.display_path = path

    def render(self) -> Content:
        self.watch_path(self.path)
        return Content(condense_path(self.display_path, self.size.width))

    def get_content_width(self, container: Size, viewport: Size) -> int:
        if self.display_path:
            return Content(self.display_path).cell_length
        else:
            return container.width



from typing import TypedDict, Literal, NotRequired

type Tag = str
"""A tag used for categorizing the agent. For example: 'open-source', 'reasoning'."""
type OS = Literal["macos", "linux", "windows", "*"]
"""An operating system identifier, or a '*" wildcard, if it is the same for all OSes."""
type Action = str
"""An action which the agent supports."""
type AgentType = Literal["coding", "chat"]
"""The type of agent. More types TBD."""
type AgentProtocol = Literal["acp"]
"""The protocol used to communicate with the agent. Currently only "acp" is supported."""


class Command(TypedDict):
    """Used to perform an action associate with an Agent."""

    description: str
    """Describes what the script will do. For example: 'Install Claude Code'."""
    command: str
    """Command to run."""
    bootstrap_uv: NotRequired[bool]
    """Bootstrap UV installer (set to `true` if the command users `uv`)."""


class Agent(TypedDict):
    """Describes an agent which Toad can connect to. Currently only Agent Client Protocol is supported.

    This information is stoed within TOML files, where the filename is the "identity" key plus the extension ".toml"

    """

    active: NotRequired[bool]
    """If `True` (default), the agent will be shown in the UI. If `False` the agent will be removed from the UI."""
    recommended: NotRequired[bool]
    """Agent is in recommended set. Set to `True` in main branch only if previously agreed with Will McGugan."""
    identity: str
    """A unique identifier for this agent. Should be a domain the agent developer owns,
    although it doesn't have to resolve to anything. Must be useable in a filename on all platforms. 
    For example: 'claude.anthropic.ai'"""
    name: str
    """The name of the agent. For example: 'Claude Code'."""
    short_name: str
    """A short name, usable on the command line. Try to make it unique. For example: 'claude'."""
    url: str
    """A URL for the agent."""
    protocol: AgentProtocol
    """The protocol used by the agent. Currently only 'acp' is supported."""
    type: "AgentType"
    """The type of the agent. Currently "coding" or "chat". More types TBD."""
    author_name: str
    """The author of the agent. For example 'Anthropic'."""
    author_url: str
    """The authors homepage. For example 'https://www.anthropic.com/'."""
    publisher_name: str
    """The publisher's name (individual or organization that wrote this data)."""
    publisher_url: str
    """The publisher's url."""
    description: str
    """A description of the agent. A few sentences max. May contain content markup (https://textual.textualize.io/guide/content/#markup) if used subtly."""
    tags: list[Tag]
    """Tags which identify the agent. Should be empty for now."""
    help: str
    """A Markdown document with additional details regarding the agent."""
    welcome: NotRequired[str]
    """A Markdown document shown to the user when the conversation starts. Should contain a welcome message and any advice on getting started."""
    run_command: dict[OS, str]
    """Command to run the agent, by OS or wildcard."""
    actions: dict[OS, dict[Action, Command]]
    """Scripts to perform actions, typically at least to install the agent."""
